/*
Created and copyright by Kwynn Buess of http://kwynn.com
written in 2011, released under GPL 2012/06/14.  See GPL (brief) terms below.

This is a Firefox AddOn / extension that counts down characters in a Facebook (! see warning) 
status update.  It counts down from the 500 (previously 420) character limit so that you see how 
many chars you have left before you hit enter.  

Please see the warning on using the word "Facebook" / the Facebook trademark: warning near the bottom.


LICENSE TERMS:

This addon / extension is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License version 3 or 
later, as published by the Free Software Foundation:
http://www.gnu.org/licenses/gpl.html

This extension is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this extension.  If not, see <http://www.gnu.org/licenses/>.
******
Kwynn note: I included the GNU GPL in a folder under the root of this XPI package.
**************************
**************************

I document in detail below.  There is also information at 

http://kwynn.com/t/1/09/counter/counter.html
and
http://kwynn.com/t/1/09/counter/supp/counter_tech.html


********** DOCUMENTATION ******************************

Once the invocations to get the addon running are done--listeners and whatnot--we must find the 
input and output fields on the Facebook (FB) page.  (Joe Hewitt's Firebug addon (http://getfirebug.com/) 
was very helpful for this.)  

Getting the I/O is complicated by the fact that the FB page doesn't really exist 
when "onload" finishes.  The JS (JavaScript) has to run next.  Therefore, it takes a number of keystrokes until 
this counter starts working.  Or it can stop and start until the page truly finishes forming.  One 
consequence is that I have to listen for keystrokes on the whole document (web page) before I can identify 
the input and output fields.  It also leads to some weirdness from the user's point of view that I describe 
on my web site.  I do have some solutions to make this better, but that's for later.

I ID the input / output by HTML "class" attributes (as of version 1.1.3 in 2012/06).  See the "findInput" and 
"findOutput" functions.

This documentation was almost entirely written for version 1.1.3 on 2012/06/14.

************
credits / acknowledgements near bottom
********
*/

/* LINE-BY-LINE DOC ************************************/

/* The equivalent of the "main" function / method --stuff that runs first--must go below the 
counter_kwynn_com object, or you'll get an "undefined" error.  So see "MAIN" below the object. The "main"
is very short.  Besides "MAIN," everything else is within the counter_kwynn_com object.
*/

// first object variables then a list of functions, then the functions themselves
var counter_kwynn_com = {

// object-scope variables below ************
MAX_CHARS_STATUS : 500, // the basic number--the number we count down from.  Not certain whether this is still the number.  
						// It was 420 until roughly spring 2011, then 500
MAX_CHARS_QUESTION: 170, // at one point FB had a "question."  Don't know details anymore.  (As of mid-2011)

BASE_COLOR: 		"#3B5998",  // blueish
OVER_LIMIT_COLOR: 	"#ED1414", // reddish

DEBUG : false,     // turns on debug mode for the error console (perhaps only works with certain FF settings)
KEY_DEBUG : false, // very low-level keystroke debugging.  

VERSION : "1.1.3", // only shown in the error console, and, even then, (perhaps?) only if certain FF settings are set.  It's not 
				   // vital that this be accurate.  It does not control upgrades, for example.

plural: true, // "character" versus "characterS" left / over the limit.  Making this an object-level var is a close call, which I 
				// may say a few more words about below.

under: true, // under or over the character limit.  Also a close call.  

FBDoc: false, // similar to "this", the actual Facebook document object / the page

status_key_cnt : 0, // I compare document level keystrokes to status message strokes - this will make more sense below
doc_key_cnt : 0,

statusInputObj : false,  // FB has both "status"es and "questions
questionInputObj: false,


// end object-scope variables *******************

/*
FUNCTION LIST

chrome_level_listener - called from outside (below) the object
loadCounter - called next
getParentDoc - used by docKeyPress
docKeyPress  - document (page) level keystroke detection
escapeHTML - mandated by AMOEs but in truth useless
doCount - status message-level listener
findOutput
findInput
addListeners - add input-level / status message listener(s) (only one at moment)
*/

// FUNCTIONS *************************

chrome_level_listener : function() // entry point from "main," which is below this object
{
    var appcontent = document.getElementById("appcontent");   // appcontent = browser, as opposed to mailer or such
    if (appcontent) appcontent.addEventListener("load", counter_kwynn_com.loadCounter, true);
}, 

loadCounter : function(event)
{	 
	if (event.target     == undefined) return;
	if (event.target.URL == undefined) return;

	if (counter_kwynn_com.DEBUG) Components.utils.reportError('Counter - url = ' + event.target.URL.substr(0,40));

	if (	
				event.target.URL == "http://www.facebook.com/" 
			|| 	event.target.URL == "https://www.facebook.com/"  
	) 
	{
		if (counter_kwynn_com.DEBUG) Components.utils.reportError('Counter - found site, otar = ' + event.target + "wrap = " + event.target.wrappedJSObject);

		counter_kwynn_com.FBDoc =  event.target.wrappedJSObject;   // define the document object used from here on out
		counter_kwynn_com.findInput(event.target.wrappedJSObject); // try to find the status box / textarea
		
		// finding the input is unlikely to work at first, so we start listening for keystrokes and keep looking for input with each key
		event.target.wrappedJSObject.addEventListener("input", counter_kwynn_com.docKeyPress, false);
	}
},


// used by docKeyPress.  When a user types ("input"s) anywhere, find the parent doc so we can starting looking for I/O.
// recursively seek parents until we find the top-level page.
getParentDoc : function (node) { 
	
	if (node == null || node == undefined) {
		if (counter_kwynn_com.DEBUG) Components.utils.reportError('Counter - null node in getParentDoc');
		return undefined;
	}

	var tempNode;
	if (node instanceof HTMLDocument) return node;
	else return counter_kwynn_com.getParentDoc(node.parentNode);
},

// use doc-level keystrokes to make this addon look for the I/O fields, if need be
docKeyPress : function (event) { 

	if (event.type == 'input') 
	{ 
		counter_kwynn_com.doc_key_cnt++;
		if (counter_kwynn_com.KEY_DEBUG) Components.utils.reportError('doc cnt = ' + counter_kwynn_com.doc_key_cnt);	
	}
	
	
	// if both document-level key counts and status-level key counts match, everything is working--we have already found input.  
	if (counter_kwynn_com.doc_key_cnt == counter_kwynn_com.status_key_cnt) { 
		if (counter_kwynn_com.KEY_DEBUG) Components.utils.reportError('count match');	
		return;
	}

	if (counter_kwynn_com.KEY_DEBUG) Components.utils.reportError('Counter - doc-level event');

	// if users are typing anywhere, we want to find the ultimate parent document / page so we can keep looking for I/O
	var FBDoc = counter_kwynn_com.getParentDoc(event.target.wrappedJSObject);

	counter_kwynn_com.findInput(FBDoc);
	counter_kwynn_com.findOutput(FBDoc);
	
	// we may have found I/O, so set these to match
	counter_kwynn_com.status_key_cnt = 0;
	counter_kwynn_com.doc_key_cnt    = 0;
},

// I only comment upon this below the object.  This was mandated by the AMOEs (defined below)
escapeHTML : function(str) { 
	
	str += ""; // forces a number (such as 420 or 500) to a string, othewise str.replace throws an error
	return str.replace(/[&"<>]/g, function (m) "&" + ({ "&": "amp", '"': "quot", "<": "lt", ">": "gt" })[m] + ";");
},


// the star of the show: actually populates the counter output; status-level listener
doCount : function (event)
{
	
	// if this function is working, the doc and status textarea inputs match--no need to look for textarea if we're in it.
	counter_kwynn_com.status_key_cnt = 0;
	counter_kwynn_com.doc_key_cnt    = 0;

	if (event.type == 'input')
	{ 
		++counter_kwynn_com.status_key_cnt;
		if (counter_kwynn_com.KEY_DEBUG) Components.utils.reportError('loc key = ' + counter_kwynn_com.status_key_cnt);	
	}

	if (counter_kwynn_com.KEY_DEBUG) Components.utils.reportError('Counter - local event');

	if (counter_kwynn_com.FBDoc == false) return; // the default value is false, as defined at top

	// the potential exception started happening in the summer of 2011 as I recall.  
	// the exception just means we lost the output again, which happens by the nature of the FB processing
	try {

		// seek the output HTML object (a certain div defined in findOutput below
		var outO = counter_kwynn_com.FBDoc.getElementById('kwynn_com_2011_count');
		if (outO == undefined || outO == null) counter_kwynn_com.findOutput();	

		if (counter_kwynn_com.statusInputObj == undefined && counter_kwynn_com.questionInputObj == undefined) 
		{ 
			if (counter_kwynn_com.DEBUG) Components.utils.reportError('Counter - inputObj undef');
			return;
		}

		// how many characters the user has typed.  I use theLen to distingush it from "length"
		var theLen = event.target.wrappedJSObject.value.length; 

		var diff;
		if (event.target.wrappedJSObject == counter_kwynn_com.statusInputObj) 
			 diff = counter_kwynn_com.MAX_CHARS_STATUS   - theLen;
		else diff = counter_kwynn_com.MAX_CHARS_QUESTION - theLen;

		
		// the "under" and "plural" object-level vars simply save processing if the actions in question are already done.  
		// this might actually matter, though.  Early versions seemed to slow things down noticably.  
		// the various fields are defined later
		if (diff < 0 && counter_kwynn_com.under == true) { 
			counter_kwynn_com.FBDoc.getElementById('kwynn_com_text_2011').innerHTML = "over the limit";
			counter_kwynn_com.FBDoc.getElementById('kwynn_com_2011_outer').style.color = counter_kwynn_com.OVER_LIMIT_COLOR; // reddish
			counter_kwynn_com.under = false;
		}
		else if (diff >= 0 && counter_kwynn_com.under == false) {
			counter_kwynn_com.FBDoc.getElementById('kwynn_com_text_2011').innerHTML = "left"; // or remaining
			counter_kwynn_com.FBDoc.getElementById('kwynn_com_2011_outer').style.color = counter_kwynn_com.BASE_COLOR; // blueish
			counter_kwynn_com.under = true;
		}

		// character versus characterS
		if ((diff == 1 || diff == -1) && counter_kwynn_com.plural) 
		{
				counter_kwynn_com.FBDoc.getElementById('kwynn_com_sing_2011').innerHTML = "";
				counter_kwynn_com.plural = false;
		} else {
			counter_kwynn_com.FBDoc.getElementById('kwynn_com_sing_2011').innerHTML = "s";
			counter_kwynn_com.plural = true;
		}

		diff = Math.abs(diff);

		counter_kwynn_com.FBDoc.getElementById('kwynn_com_2011_count').innerHTML = counter_kwynn_com.escapeHTML(diff);

		if (counter_kwynn_com.KEY_DEBUG) Components.utils.reportError("count HTML = " + counter_kwynn_com.FBDoc.getElementById('kwynn_com_2011_count').innerHTML);
	} // end try
	// in the event we lost output, exception runs, and we look again
	catch(e) { 
		if (counter_kwynn_com.KEY_DEBUG) Components.utils.reportError('Counter doCount() exception.  e = ' + e);
		counter_kwynn_com.findOutput();
	}
}, // end doCount

// find the appropriate FB HTML object for output, and add our output
findOutput: function() {

	if (counter_kwynn_com.DEBUG) Components.utils.reportError('counter - in findOutput()');

	// if everything's defined, we don't need to be here
	if (counter_kwynn_com.FBDoc != undefined)
		if (	counter_kwynn_com.FBDoc.getElementById('kwynn_com_2011_count') != undefined && 
				counter_kwynn_com.FBDoc.getElementById('kwynn_com_2011_count') != null
		) return;
	
	var diva = counter_kwynn_com.FBDoc.getElementsByTagName('div'); // div array
	
	if (counter_kwynn_com.DEBUG) Components.utils.reportError('counter - div count = ' + diva.length);

	var count = 0;
	for (var i=0; i < diva.length; i++) 
	{
		// by observation, only our output field has these HTML class attribute combinations
		if (
			(		diva[i].className.search(/showOnceInteracted/) >= 0
				&& 	diva[i].className.search(/uiComposerMessageBoxControls/) >= 0
			) || 	diva[i].className.search(/attachmentBarArea/) >= 0
		) 
		{ 
			if (counter_kwynn_com.DEBUG) Components.utils.reportError('counter - found output div - v');
	
			var newP = counter_kwynn_com.FBDoc.createElement('p');
			newP.style.display = 'inline';
						
			/* the kwynn_com_2011 is to distinguish the new fields from anything another addOn writer might define. 
				outer = the output that will turn colors / outermost useful span
				count = the numerical count
				sing  = singual versus plural
				left  = "left" (remaining) or "over the limit"	*/
			newP.innerHTML = 
				"&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; " +
				"<span style='color: " + counter_kwynn_com.BASE_COLOR + "; font-weight: bold' id='kwynn_com_2011_outer'>" + 
						"<span id='kwynn_com_2011_count' >" + 
							counter_kwynn_com.escapeHTML(counter_kwynn_com.MAX_CHARS_STATUS) + 
						"</span> " + 
						"character" + "<span id='kwynn_com_sing_2011'>s</span>" + " " +
						"<span id='kwynn_com_text_2011'>left</span>" +
				 "</span>" 
				;

			diva[i].appendChild(newP);

			if (counter_kwynn_com.DEBUG) Components.utils.reportError('counter - found output div');

			return true;
			
		} // end if proper output
	} // end loop

	return false;
}, // end findOutput

findInput : function(docIN) { // input = status message box / textarea HTML field

	if (docIN == undefined) if (counter_kwynn_com.DEBUG) Components.utils.reportError('Counter - undefined doc sent to findInput');
	if (docIN == undefined) return false;

	var texta = docIN.getElementsByTagName('textarea'); // text(area) array

	if (counter_kwynn_com.DEBUG) Components.utils.reportError('Counter - in findInput');

	for (var i=0; i < texta.length; i++) 
	{

		if (counter_kwynn_com.DEBUG) Components.utils.reportError('Counter - in fs loop');
	
		var tempObj = texta[i];
		var taClass = texta[i].className; // textarea class name
		
		if (taClass == undefined) continue;
		
		if (counter_kwynn_com.DEBUG) Components.utils.reportError('ta class name = ' + taClass);

		// by observation, here is how to define our input
		if (	
			(
				taClass.search(/uiTextareaAutogrow/) >= 0 && 
				taClass.search(/input/) >= 0 &&
				taClass.search(/mentionsTextarea/) >= 0 &&
				taClass.search(/textInput/) >= 0
			) 
			||
			(
				taClass.search(/uiTextareaNoResize/) >= 0 &&
				taClass.search(/uiTextareaAutogrow/) >= 0 &&
				taClass.search(/questionInput/)      >= 0 &&
				taClass.search(/fluidInput/)         >= 0
			)
		) {

			if (counter_kwynn_com.DEBUG) Components.utils.reportError('Counter - found input div - ID = ' + tempObj + ' - v' + this.VERSION);

			var maxCountSpan = docIN.getElementById('kwynn_com_2011_count');

			// status versus FB "question"
			if (taClass.search(/textInput/) >= 0) {
				counter_kwynn_com.statusInputObj = tempObj;
				if (maxCountSpan != undefined) maxCountSpan.innerHMTL = counter_kwynn_com.MAX_CHARS_STATUS;
			}
			else {
				counter_kwynn_com.questionInputObj  = tempObj;
				if (maxCountSpan != undefined) maxCountSpan.innerHMTL = counter_kwynn_com.MAX_CHARS_QUESTION;

			}

			counter_kwynn_com.FBDoc = docIN;
			counter_kwynn_com.addListeners(tempObj);
			counter_kwynn_com.findOutput();

			return true;
		} // if className matches
	} // end for loop for looping through textareas

	return false;

}, // end function

// originally there were more listeners, but I'll leave this a function
addListeners : function(object)
{
	object.addEventListener('input', counter_kwynn_com.doCount, false);
}

}; // end object

// ********************* MAIN ***************************

if (counter_kwynn_com.DEBUG) Components.utils.reportError('Counter - hello world - very first - v' + counter_kwynn_com.VERSION);

window.addEventListener("load", counter_kwynn_com.chrome_level_listener, false);
// END PROGRAM - ONLY DOC BELOW **************************************

/**************** escapeHTML DOC ************************/

/* The escapeHTML function just below comes almost verbatim from the AMO (addons.mozilla.org) Editors in reponse to the 
submission of this AddOn for "full review."  You can see this function at 
https://wiki.mozilla.org/AMO:Editors/MiscCannedResponses  */

// The following line is exactly what the AMO Editors sent.  
//   function escapeHTML(str) str.replace(/[&"<>]/g, function (m) "&" + ({ "&": "amp", '"': "quot", "<": "lt", ">": "gt" })[m] + ";");

// I had to modify it just slightly for whatever this syntax is called -- class / object syntax or whatever:
// ("name : function" syntax rather than "function name"

// I am annoyed by the escape function because there is no user data involved that can cause a cross scripting attack or whatnot


/************* CREDITS ******************

Stephen L. Brown of BizGenie.net gave me the idea, although the idea had been floating around for at least (roughly) a year, and 
others have done such a counter.

Stephen also introduced me to the Firebug Firefox AddOn by Joe Hewitt.  That speeded development quite a bit.  
Stephen also also (sic) dictated the style--color and placement of my output box.

My sister, Kyleigh Buess, helped debug 0.2.5 so I could improve to 0.3.0.  

Will McSweeney of BizGenie.net pointed out that it's keyup, not keypress.  (This is moot now because I'm using "input," but he 
still helped.)

As mentioned, Joe Hewitt's Firebug addon:
http://getfirebug.com/ 
was very helpful for id'ing my I/O.  

The first versions of this were not addons but "Greasemonkey" scripts.  So, credit to the Greasemonkey FF AddOn:
https://addons.mozilla.org/en-US/firefox/addon/greasemonkey/
by Anthony Lieuallen, Aaron Boodman, Johan Sundstrm
*/

/* PRIOR (?) WORK / prior art ****************************************

I don't know for sure, but I hear that the "Social Fixer (previously Better Facebook)" addon by Matt Kruse does the same thing.  By the time
I found out about this possibility, I was already done.  

*/

/******************** ORIGIN HISTORICAL NOTE ***************************
Version 0.0.4 - first stable version, 3:00am
very first version created March 17, 2011, ca. 12:30am EDT (00:30, USA, GMT -4)
*/

/********************* WARNING ON THE FACEBOOK TRADEMARK / SOME PROJECT HISTORY *************************/
/*
At least one version of this AddOn was approved as a fully reviewed Firefox addon by the AMOEs (Addons.Mozilla.org Editors).  
A previous version was briefly listed on AMO.  Very soon after, though, Facebook found it and declared it trademark infringement.  
It was taken down.  I understand that "Better Facebook" had the same problem, and probably many others.  Beware of putting the word 
"Facebook" back in user-visible position.  I can send you Facebook's emails to me if you want.  I'm pretty easy to find.  
*/
